from typing import Iterable, List, Any, Tuple
from functools import singledispatch, reduce
from itertools import groupby


def __next_current(current: str, n: str):
    if current != "":
        return f"{current}.{n}"
    return n


@singledispatch
def flat(value, current: str = "") -> Iterable[tuple]:
    yield current, value


@flat.register(dict)
def _(d: dict, current: str = "") -> Iterable[tuple]:
    for k, v in d.items():
        nc = __next_current(current, k)
        yield from flat(v, nc)


@flat.register(list)
def _(l: list, current: str = "") -> Iterable[tuple]:
    for i, v in enumerate(l):
        nc = __next_current(current, f"[{i}]")
        yield from flat(v, nc)


def __part_type(part: str) -> str | int:
    if part.startswith("[") and part.endswith("]"):
        return int(part[1:-1])
    return part


def __nest_recursion(data, part_value):
    parts, value = part_value
    if isinstance(parts, list) and len(parts) == 1:
        if isinstance(parts[0], int) and len(data) <= parts[0]:
            data.extend([None] * (parts[0] - len(data) + 1))
        data[parts[0]] = value
    elif isinstance(parts, list) and isinstance(parts[0], str):
        if parts[0] not in data:
            data[parts[0]] = {}
        data[parts[0]] = __nest_recursion(data[parts[0]], (parts[1:], value))
    elif isinstance(parts[0], int):
        if len(data) <= parts[0]:
            data.extend([None] * (parts[0] - len(data) + 1))
        data[parts[0]] = __nest_recursion(data[parts[0]], (parts[1:], value))
    else:
        raise NotImplementedError
    return data


def nest(items: List[Tuple[str, Any]]): 
    keys = [x for x, _ in items]
    level_parts = [[__part_type(p) for p in x.split(".")] for x in keys]
    values = [x for _, x in items]

    # type check
    types = set(map(type, [x[0] for x in level_parts]))
    if len(types) > 1:
        raise ValueError("Cannot nest mixed types")
    level_type = types.pop()
    type_values = list(zip(level_parts, values))

    if level_type == str:
        return reduce(__nest_recursion, type_values, {})
    elif level_type == int:
        return reduce(__nest_recursion, type_values, [])
    raise NotImplementedError


__all__ = ["flat", "nest"]



import pytest

from dictionary_search import nest


def test_mixing_types():
    with pytest.raises(ValueError):
        d = [("a", 1), ("[0]", 2)]
        nest(d)


def test_simple_nest():
    base = {"a": 1, "b": 2}
    d = list({"a": 1, "b": 2}.items())
    assert nest(d) == base 


def test_nest_prop():
    d = list({"a.b": 1, "a.c": 2}.items())
    assert nest(d) == {"a": {"b": 1, "c": 2}}


def test_mixed_nesting():
    d = list({"a.b": 1, "a.c": 2, "d": 3}.items())
    assert nest(d) == {"a": {"b": 1, "c": 2}, "d": 3}


def test_list_nesting():
    d = [("a.b", 1), ("a.c", 2), ("d", 3)]
    assert nest(d) == {"a": {"b": 1, "c": 2}, "d": 3}


def test_index_nesting():
    d = [("[0]", 1), ("[1]", 2), ("[2]", 3)]
    assert nest(d) == [1, 2, 3]



from dictionary_search import search_first

def test_search_one():
    d = {
        "a": {
            "b": {
                "c": "C value 1"
            }
        },
        "e": "f",
        "h": [
            {
                "i": "j"
            },
            {
                "c": "C value 2"
            }
        ]
    }

    res = search_first(d, "c")

    assert res[0] == "a.b.c"
    assert res[1] == "C value 1"




from typing import Tuple, Set, List


def _search(current, target, path) -> Tuple[str, object]:
    if isinstance(current, dict):
        if target in current:
            yield (*path, target), current[target]

        for x, y in current.items():
            for res in _search(y, target, (*path, x)):
                yield res
    elif isinstance(current, list):
        for i, item in enumerate(current):
            for res in _search(item, target, (*path, f"[{i}]")):
                yield res


def search_first(tree: dict,
                 target: str,
                 ancestors: Set[str] = None,
                 path_as_string: bool = True) -> Tuple[str, str]:
    ancestors = ancestors or set([])

    for (path, element) in _search(tree, target, tuple()):
        if ancestors <= set(path):

            if path_as_string:
                _path = ".".join(path)
            else:
                _path = path

            return _path, element


def search_all(tree: dict,
               target: str,
               ancestors: Set[str] = None,
               path_as_string: bool = True) -> List[Tuple[str, str]]:
    ancestors = ancestors or set([])

    res = list()

    for (path, element) in _search(tree, target, tuple()):
        if ancestors <= set(path):
            if path_as_string:
                _path = ".".join(path)
            else:
                _path = path

            res.append((_path, element))

    return res

__all__ = ["search_first", "search_all"]
